/*
 * Copyright (C) 2024 sillysagiri
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include <cassert>
#include <cstdint>
#include <cstdlib>
#include <filesystem>
#include <fstream>
#include <stdexcept>
#include <format>
#include <iostream>
#include <vector>

#define STB_IMAGE_IMPLEMENTATION
#define STB_IMAGE_WRITE_IMPLEMENTATION

#include <argparse/argparse.hpp>
#include "libimagequant.h"
#include "header.hpp"
#include "Timer.h"
#include "fastlz.h"
#include "murmurhash.h"

namespace fs = std::filesystem;

int main(int argc, char **argv)
{
	srand(time(0));
	argparse::ArgumentParser program("sillyimage", "1.3");
	Options opt;

	// Register argument
	{
		program.add_argument("-i", "--input")
			.help("input files")
			.append()
			.required()
			.store_into(opt.path_input);

		program.add_argument("-o", "--out")
			.help("output directory")
			.required()
			.store_into(opt.path_output);

		program.add_argument("-f", "--format")
			.help("texture format { rgb256, rgb16, indexed4, indexed16, indexed256, indexed32a3, indexed8a5 }")
			.default_value<std::string>("rgb16")
			.choices("rgb256", "rgb16", "indexed4", "indexed16", "indexed256", "indexed32a3", "indexed8a5")
			.nargs(1)
			.store_into(opt.format);
	}

	// Parse argument
	{
		try
		{
			program.parse_args(argc, argv);

			// input list
			for (auto &i : opt.path_input)
				if (i.empty() || !fs::exists(i)) 
					throw std::runtime_error(std::format("file doesnt exsist: \"{}\"", i));

			// output directory
			if (opt.path_output.empty() || !fs::is_directory(opt.path_output))
				throw std::runtime_error(std::format("invalid output directory: \"{}\"", opt.path_output));

		}
		catch(const std::exception &e)
		{
			std::cerr << std::format("Error: {}\n\n", e.what()) << program << "\n";
			return 1;
		}
	}

	// Convert
	try {

		if (opt.format == "rgb256")
		{
			for(auto &input : opt.path_input)
			{
				Image img;
				img.Load(input);

				Metadata meta;
				meta.format = Format::Format_RGB_256;
				meta.width = img.width;
				meta.height = img.height;
				meta.palette_count = 0;
				meta.palette_hash = 0;
				meta.original_size = img.width*img.height*img.comp;

				uint8_t compress[uint64_t(meta.original_size*1.5)];
				meta.compress_size = fastlz_compress_level(1, img.data, meta.original_size, compress);

				auto output = fs::path(opt.path_output) / std::format("{}.sillyimg", fs::path(input).stem().string());
				std::ofstream out(output, std::ios::binary);
				out.write(reinterpret_cast<const char*>(&meta), sizeof(meta));
				out.write(reinterpret_cast<const char*>(compress), meta.compress_size);
				out.close();
			}
		}

		if (opt.format == "rgb16")
		{
			for(auto &input : opt.path_input)
			{
				Image img;
				img.Load(input);
				uint16_t img16[img.width*img.height];

				// convert into RGB16 color
				for (int i=0; i<img.width*img.height; i++)
				{
					uint8_t a = img.data[i*4+3];
					uint8_t r = img.data[i*4+0];
					uint8_t g = img.data[i*4+1];
					uint8_t b = img.data[i*4+2];

					img16[i] = RGB16(a, r, g, b);
				}

				Metadata meta;
				meta.format = Format::Format_RGB_16;
				meta.width = img.width;
				meta.height = img.height;
				meta.palette_count = 0;
				meta.palette_hash = 0;
				meta.original_size = img.width*img.height*2;

				uint8_t compress[uint64_t(meta.original_size*1.5)];
				meta.compress_size = fastlz_compress_level(1, img16, meta.original_size, compress);

				auto output = fs::path(opt.path_output) / std::format("{}.sillyimg", fs::path(input).stem().string());
				std::ofstream out(output, std::ios::binary);
				out.write(reinterpret_cast<const char*>(&meta), sizeof(meta));
				out.write(reinterpret_cast<const char*>(compress), meta.compress_size);
				out.close();
			}
		}

		if (opt.format == "indexed4" || opt.format == "indexed16" || opt.format == "indexed256")
		{
			uint16_t numcol;
			uint16_t bpp;

			if (opt.format == "indexed4")
			{
				numcol = 4;
				bpp = 2;
			}
			if (opt.format == "indexed16")
			{
				numcol = 16;
				bpp = 4;
			}
			if (opt.format == "indexed256")
			{
				numcol = 256;
				bpp = 8;
			}

			std::vector<Color> palette;
			std::vector<Image> images(opt.path_input.size());
			std::vector<IndexedImage> indexes(opt.path_input.size());
			
			for (int i=0; i<images.size(); i++)
				images[i].Load(opt.path_input[i]);

			QuantizeImage(numcol, images, palette, indexes);
			uint16_t palette16[palette.size()];
			for (const auto &pal : palette)
				uint16_t rgb16 = RGB16(pal.a, pal.r, pal.g, pal.b);

			for (int i=0; i<indexes.size(); i++)
			{
				uint32_t length = indexes[i].width*indexes[i].height;
				if (bpp < 8) BitPacking(bpp, indexes[i], length);

				Metadata meta;
				meta.format = Format::Format_RGB_16;
				meta.width = indexes[i].width;
				meta.height = indexes[i].height;
				meta.palette_count = numcol;
				meta.palette_hash = murmurhash(reinterpret_cast<const char*>(palette16), sizeof(uint16_t) * palette.size(), 1);
				meta.original_size = length;

				uint8_t compress[uint64_t(meta.original_size*1.5)];
				meta.compress_size = fastlz_compress_level(1, indexes[i].data, meta.original_size, compress);

				auto output = fs::path(opt.path_output) / std::format("{}.sillyimg", fs::path(opt.path_input[i]).stem().string());
				std::ofstream out(output, std::ios::binary);
				
				out.write(reinterpret_cast<const char*>(&meta), sizeof(meta));
				out.write(reinterpret_cast<const char*>(palette16), sizeof(uint16_t) * palette.size());
				out.write(reinterpret_cast<const char*>(compress), meta.compress_size);
				out.close();
			}
		}

		if (opt.format == "indexed32a3" || opt.format == "indexed8a5")
			throw std::runtime_error("format not supported yet!");

	}
	catch(const std::exception &e)
	{
		std::cerr << std::format("Exception caught: {}\n", e.what());
		return 1;
	}
	catch (...)
	{
		std::cerr << std::format("An unknown exception occurred\n");
		return 1;
	}

	std::cout  << "done\n";

	return 0;
}

// // ----

uint16_t RGB16(uint8_t a, uint8_t r, uint8_t g, uint8_t b)
{
	// TODO: alpha threshold
	uint16_t a1 = a & 0x01;
	uint16_t r5 = (r >> 3) & 0x1F;
	uint16_t g5 = (g >> 3) & 0x1F;
	uint16_t b5 = (b >> 3) & 0x1F;

	return (a1 << 15) | (r5) | (g5 << 5) | (b5 << 10);
}

void QuantizeImage(const int numColor, std::vector<Image> &images, std::vector<Color> &palette, std::vector<IndexedImage> &indexes)
{
	liq_attr *liq_attr;
	liq_result *liq_result;
	liq_histogram *liq_histogram;
	const liq_palette *liq_palette;

	liq_attr = liq_attr_create();
	liq_histogram = liq_histogram_create(liq_attr);

	liq_set_speed(liq_attr, 1);
	liq_set_max_colors(liq_attr, numColor);

	std::vector<liq_image*> liq_images(images.size());

	for (int i=0; i<images.size(); i++)
	{
		liq_images[i] = liq_image_create_rgba(liq_attr, images[i].data, images[i].width, images[i].height, 0); 
		liq_histogram_add_image(liq_histogram, liq_attr, liq_images[i]);
	}

	auto liq_err = liq_histogram_quantize(liq_histogram, liq_attr, &liq_result);
	liq_set_dithering_level(liq_result, 0);
	liq_palette = liq_get_palette(liq_result);
	assert(liq_err == LIQ_OK);

	palette.resize(liq_palette->count);
	for(int i=0; i<liq_palette->count; i++)
		palette[i] = {
			liq_palette->entries[i].a,
			liq_palette->entries[i].r,
			liq_palette->entries[i].g,
			liq_palette->entries[i].b };

	for (int i=0; i<images.size(); i++)
	{
		indexes[i].Load(images[i].width, images[i].height);
		liq_err = liq_write_remapped_image(liq_result, liq_images[i], indexes[i].data, indexes[i].width*indexes[i].height);
		if (liq_err != LIQ_OK) throw std::runtime_error(std::format("{} {} {}", static_cast<int>(liq_err), indexes[i].width, indexes[i].height));
	}

	for (auto &i : liq_images)
		liq_image_destroy(i);

	liq_attr_destroy(liq_attr);
	liq_histogram_destroy(liq_histogram);
	liq_result_destroy(liq_result);
}

void BitPacking(const uint8_t bpp, IndexedImage &img, uint32_t &length)
{
	length = ((img.width*img.height) * bpp) / 8;
	uint8_t *buffer = new uint8_t[length]();
	uint32_t bufferPos = 0;
	uint8_t bitpos = 0;

	for (int i2=0; i2<(img.width*img.height); i2++)
	{
		buffer[bufferPos] |= img.data[i2] << bitpos;
		bitpos += bpp;

		if (bitpos == 8)
		{
			bitpos = 0;
			bufferPos++;
		}
	}

	delete[] img.data;
	img.data = buffer;
}

// // ----

Image::Image(const std::filesystem::path &path)
{
	Load(path);
}

Image::~Image()
{
	stbi_image_free(data);
}

void Image::Load(const std::filesystem::path &path)
{
	data = stbi_load(path.c_str(), &width, &height, &comp, 4);
	if (data == nullptr) throw std::runtime_error(stbi_failure_reason());
}

// // ----

IndexedImage::IndexedImage(int width, int height)
	: width(width), height(height), data(nullptr)
{
	Load(width, height);
}

IndexedImage::~IndexedImage()
{
	delete[] data; 
}

void IndexedImage::Load(int width, int height)
{
	this->width = width;
	this->height = height;

	data = new uint8_t[width*height];
	if (data == nullptr) throw std::runtime_error("not enough memory");
}